#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import struct

from .freq import EnumLruMapping, YajbeEncoderEnumConfig


def int_bytes_width(v: int) -> int:
    return (v.bit_length() + 7) // 8 if v != 0 else 1


class FieldNameWriter:
    def __init__(self, encoder, initial_field_names: list[str] = None) -> None:
        self._encoder = encoder
        self._indexed_map = {}
        self._last_key = b''
        if initial_field_names:
            for i, name in enumerate(initial_field_names[:65819]):
                self._indexed_map[name] = i

    def encode_string(self, key: str) -> None:
        utf8 = key.encode('utf-8')

        index = self._indexed_map.get(key)
        if index is not None:
            self._write_indexed_field_name(index)
            self._last_key = utf8
            return

        if self._last_key and len(utf8) > 4:
            prefix = min(0xff, self._prefix(utf8))
            suffix = self._suffix(utf8, prefix)

            if suffix > 2:
                self._write_prefix_suffix(utf8, prefix, min(0xff, suffix))
            elif prefix > 2:
                self._write_prefix(utf8, prefix)
            else:
                self._write_full_field_name(utf8)
        else:
            self._write_full_field_name(utf8)

        if len(self._indexed_map) < 65819:
            self._indexed_map[key] = len(self._indexed_map)
        self._last_key = utf8

    def _write_full_field_name(self, utf8: bytes):
        # 100----- Full Field Name (0-29 length - 1, 30 1b-len, 31 2b-len)
        self._write_length(0b100_00000, len(utf8))
        self._encoder._write_bytes(utf8)

    def _write_indexed_field_name(self, index: int):
        # 101----- Field Offset (0-29 field, 30 1b-len, 31 2b-len)
        self._write_length(0b101_00000, index)

    def _write_prefix(self, utf8: bytes, prefix: int):
        # 110----- Prefix (1byte prefix, 0-29 length - 1, 30 1b-len, 31 2b-len)
        length = len(utf8) - prefix
        self._write_length(0b110_00000, length)
        self._encoder._write_byte(prefix)
        self._encoder._write_bytes(utf8[prefix:])

    def _write_prefix_suffix(self, utf8: bytes, prefix: int, suffix: int):
        # 111----- Prefix/Suffix (1byte prefix, 1byte suffix, 0-29 length - 1, 30 1b-len, 31 2b-len)
        length = len(utf8) - prefix - suffix
        self._write_length(0b111_00000, length)
        self._encoder._write_byte(prefix)
        self._encoder._write_byte(suffix)
        self._encoder._write_bytes(utf8[prefix:prefix + length])

    def _write_length(self, head: int, length: int):
        if length < 30:
            self._encoder._write_byte(head | length)
        elif length <= 284:
            # 30 + 1byte = 284
            self._encoder._write_byte(head | 0b11110)
            self._encoder._write_byte((length - 29) & 0xff)
        elif length <= 65819:
            # 31 + 2byte = 65819
            self._encoder._write_byte(head | 0b11111)
            self._encoder._write_byte((length - 284) // 256)
            self._encoder._write_byte((length - 284) & 255)
        else:
            raise Exception("unexpected too many field names: %s" % length)

    def _prefix(self, key: bytes) -> int:
        last_key = self._last_key
        min_len = min(len(last_key), len(key))
        for i in range(min_len):
            if last_key[i] != key[i]:
                return i
        return min_len

    def _suffix(self, key: bytes, key_prefix: int) -> int:
        last_key = self._last_key
        last_key_len = len(last_key)
        key_len = len(key) - key_prefix
        min_len = min(len(last_key), key_len)
        for i in range(1, min_len + 1):
            if (last_key[last_key_len - i] & 0xff) != (key[key_prefix + (key_len - i)] & 0xff):
                return i - 1
        return min_len

from dataclasses import is_dataclass, asdict as dataclass_as_dict
class YajbeEncoder:
    def __init__(self, stream: io.BufferedIOBase, initial_field_names: list[str] = None, enum_config: YajbeEncoderEnumConfig = None) -> None:
        self._stream = stream
        self._field_name_writer = FieldNameWriter(self, initial_field_names)
        self._enum_config = enum_config
        self._enum_mapping = None
        self._types_map = {
            bool: self.encode_bool,
            int: self.encode_int,
            float: self.encode_float,
            bytes: self.encode_bytes,
            bytearray: self.encode_bytes,
            str: self.encode_string,
            list: self.encode_array,
            tuple: self.encode_array,
            set: self.encode_array,
            dict: self.encode_object,
        }

    def encode_item(self, item):
        if item is None:
            return self.encode_null()

        item_type = type(item)
        encoder = self._types_map.get(item_type)
        if encoder is None:
            if is_dataclass(item):
                item = dataclass_as_dict(item)
                return self.encode_object(item)
            raise Exception('unsupported type %s: %s' % (item_type, item))
        return encoder(item)

    def encode_null(self):
        self._write_byte(0)

    def encode_bool(self, value: bool):
        self._write_byte(3 if value else 2)

    def _encode_positive_int(self, value: int) -> None:
        if value <= 24:
            self._write_byte(0b010_00000 | (value - 1))
        else:
            value -= 25
            nbytes = int_bytes_width(value)
            self._write_byte(0b010_00000 | (23 + nbytes))
            self._write_uint(value, nbytes)

    def _encode_negative_int(self, value: int) -> None:
        value = -value
        if value <= 23:
            self._write_byte(0b011_00000 | value)
        else:
            value -= 24
            nbytes = int_bytes_width(value)
            self._write_byte(0b011_00000 | (23 + nbytes))
            self._write_uint(value, nbytes)

    def encode_int(self, value: int) -> None:
        if value > 0:
            self._encode_positive_int(value)
        else:
            self._encode_negative_int(value)

    def encode_float(self, value: float) -> None:
        data = struct.pack('<d', value)
        self._write_byte(0b00000_110)
        self._stream.write(data)

    def encode_string(self, value: str) -> None:
        if self._enum_config and self.encode_enum(value):
            return

        utf8 = value.encode('utf-8')
        self._write_length(0b11_000000, 59, len(utf8))
        self._stream.write(utf8)

    def encode_enum(self, text: str) -> bool:
        if self._enum_mapping is None:
            self.new_enum_mapping()

        index = self._enum_mapping.add(text)
        if index < 0: return False

        if index <= 0xff:
            self._write_byte(0b00001001)
            self._write_byte(index)
        elif index <= 0xffff:
            self._write_byte(0b00001010)
            self._write_uint(index, 2)
        else:
            raise Exception("enum index too large " + index)
        return True

    def new_enum_mapping(self) -> None:
        config = self._enum_config
        match config.TYPE:
            case 'LRU':
                self._enum_mapping = EnumLruMapping(config.lru_size, config.min_freq)

                self._write_byte(0b00001000)
                self._write_byte(config.lru_size.bit_length() - 6)
                self._write_byte(config.min_freq - 1)
                return

    def encode_bytes(self, value: bytes) -> None:
        self._write_length(0b10_000000, 59, len(value))
        self._stream.write(value)

    def encode_object(self, obj: dict) -> None:
        keys = obj.keys()
        self._write_length(0b0011_0000, 10, len(keys))
        for key in keys:
            self._field_name_writer.encode_string(key)
            self.encode_item(obj[key])

    def encode_array(self, array: list) -> None:
        self._write_length(0b0010_0000, 10, len(array))
        for v in array:
            self.encode_item(v)

    def _write_length(self, head: int, inline_max: int, length: int) -> None:
        if length <= inline_max:
            self._write_byte(head | length)
        else:
            delta_length = length - inline_max
            nbytes = int_bytes_width(delta_length)
            self._write_byte(head | (inline_max + nbytes))
            self._write_uint(delta_length, nbytes)

    def _write_byte(self, v: int) -> None:
        self._stream.write(bytes([v & 0xff]))

    def _write_bytes(self, value: bytes) -> None:
        self._stream.write(value)

    def _write_uint(self, value: int, width: int) -> None:
        buf = bytearray(width)
        for i in range(width):
            buf[i] = (value >> (i << 3)) & 0xff
        self._stream.write(buf)


def encode_to_stream(stream: io.BufferedIOBase, value, initial_field_names: list[str] = None, enum_config: YajbeEncoderEnumConfig = None) -> None:
    decoder = YajbeEncoder(stream, initial_field_names, enum_config)
    decoder.encode_item(value)


def encode_as_bytes(value, initial_field_names: list[str] = None, enum_config: YajbeEncoderEnumConfig = None) -> bytes:
    with io.BytesIO() as stream:
        encode_to_stream(stream, value, initial_field_names, enum_config)
        return stream.getvalue()



#!/usr/bin/env python3

from yajbe import YajbeEnumLruConfig
import yajbe

# simple usage: encode and decode from bytes
enc = yajbe.encode_as_bytes({'a': 10, 'b': ['hello', 10]})
dec = yajbe.decode_bytes(enc)
print(dec)

# options: identify common strings and avoid writing them every time
enc = yajbe.encode_as_bytes([{'a': "foooo"}, {'a': "foooo"}, {'a': "foooo"}, {'a': "foooo"}], enum_config=YajbeEnumLruConfig(256, 4))
dec = yajbe.decode_bytes(enc)
print(dec)

# encode directly to a stream
with open('test.yajbe', 'wb') as fd:
    yajbe.encode_to_stream(fd, {'a': 10, 'b': ['hello', 10]})

# decode directly from a stream
with open('test.yajbe', 'rb') as fd:
    print(yajbe.decode_stream(fd))  # {'a': 10, 'b': ['hello', 10]}

from dataclasses import dataclass, asdict as dataclass_as_dict

@dataclass
class Bar:
    x: int
    y: str

@dataclass
class Foo:
    a: int
    b: Bar

f = Foo(10, Bar(5, 'foo'))
print(yajbe.encode_as_bytes(f))



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from dataclasses import dataclass

MIN_ENUM_STRING_LENGTH = 3
MAX_INDEX_LENGTH = 0xffff

class YajbeEncoderEnumConfig:
    TYPE = None

class YajbeEnumLruConfig(YajbeEncoderEnumConfig):
    TYPE = 'LRU'

    def __init__(self, lru_size, min_freq):
        self.lru_size = lru_size
        self.min_freq = min_freq

@dataclass
class EnumLruItem:
  key: str
  freq: int

class EnumLruMapping:
    def __init__(self, lru_size, min_freq) -> None:
        self.lru_size = lru_size
        self.min_freq = min_freq
        self._lru = deque(maxlen=lru_size)
        self._lru_map = {}
        self._index_map = {}
        self._indexed = []

    def get(self, index: int) -> str:
        return self._indexed[index]

    def add(self, key: str) -> int:
        if len(key) < MIN_ENUM_STRING_LENGTH: return -1

        key_index = self._index_map.get(key)
        if key_index is not None: return key_index

        if len(self._indexed) == MAX_INDEX_LENGTH:
            return -1

        item = self._lru_map.get(key)
        if item is None:
            if len(self._lru) == self.lru_size:
                evict_item = self._lru.pop()
                del self._lru_map[evict_item.key]

            new_item = EnumLruItem(key, 1)
            self._lru.appendleft(new_item)
            self._lru_map[new_item.key] = new_item
            return -1

        self._lru.remove(item)
        item.freq += 1
        if item.freq >= self.min_freq:
            new_index = len(self._index_map)
            self._indexed.append(item.key)
            self._index_map[item.key] = new_index
            #print('add to index map', len(self._index_map), item.key, new_index)
            del self._lru_map[item.key]
        else:
          self._lru.appendleft(item)
        return -1

if __name__ == '__main__':
    data = ["2022-12", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "west-mercia", "west-midlands", "west-yorkshire", "2022-11", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "west-mercia", "west-midlands", "west-yorkshire", "2022-10", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "west-mercia", "west-midlands", "west-yorkshire", "2022-09", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "west-mercia", "west-midlands", "west-yorkshire", "2022-08", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2022-07", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2022-06", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2022-05", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2022-04", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2022-03", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2022-02", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2022-01", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-12", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-11", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-10", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-09", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-08", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-07", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-06", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-05", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-04", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-03", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-02", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2021-01", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-12", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-11", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-10", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-09", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-08", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-07", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-06", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-05", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-04", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-03", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-02", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire", "2020-01", "avon-and-somerset", "bedfordshire", "btp", "cambridgeshire", "cheshire", "city-of-london", "cleveland", "cumbria", "derbyshire", "devon-and-cornwall", "dorset", "durham", "dyfed-powys", "essex", "gloucestershire", "gwent", "hampshire", "hertfordshire", "humberside", "kent", "lancashire", "leicestershire", "lincolnshire", "merseyside", "metropolitan", "norfolk", "north-wales", "north-yorkshire", "northamptonshire", "northumbria", "nottinghamshire", "south-wales", "south-yorkshire", "staffordshire", "suffolk", "surrey", "sussex", "thames-valley", "warwickshire", "west-mercia", "west-midlands", "west-yorkshire", "wiltshire"]

    freq = EnumLruMapping(256, 3)
    for key in data:
        index = freq.add(key)
        if index is not None:
            print(index, key)
    print('bang', freq._index_map)
    #print('bang', freq._lru)
    #print('bang', freq._lru_map)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License") you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import random

from yajbe import encode_as_bytes, decode_bytes


class TestYajbeArray(unittest.TestCase):
    def test_simple(self):
        self.assertDecode("20", [])
        self.assertDecode("2f01", [0] * 0)
        self.assertDecode("2f6001", [0] * 1)
        self.assertDecode("2f606001", [0] * 2)
        self.assertDecode("2f4001", [1])
        self.assertDecode("2f414101", [2, 2])

        self.assertEncodeDecode([0] * 0, "20")
        self.assertEncodeDecode([1], "2140")
        self.assertEncodeDecode([2, 2], "224141")
        self.assertEncodeDecode([0] * 10, "2a60606060606060606060")
        self.assertEncodeDecode([0] * 11, "2b016060606060606060606060")
        self.assertEncodeDecode([0] * 0xff, "2bf5" + "60" * 0xff)
        self.assertEncodeDecode([0] * 265, "2bff" + "60" * 265)
        self.assertEncodeDecode([0] * 0xffff, "2cf5ff" + "60" * 0xffff)
        #self.assertEncodeDecode([0] * 0xffffff, "2df5ffff" + "60" * 0xffffff) # too slow
        #self.assertEncodeDecode([0] * 0x1fffffff, "2e1fffff" + "60" * 0x1fffffff) # too slow

        self.assertEncodeDecode([None] * 0, "20")
        self.assertEncodeDecode([None] * 1, "2100")
        self.assertEncodeDecode([""], "21c0")
        self.assertEncodeDecode(["a"], "21c161")
        self.assertEncodeDecode([None] * 0xff, "2bf5" + "00" * 0xff)
        self.assertEncodeDecode([None] * 265, "2bff" + "00" * 265)
        self.assertEncodeDecode([None] * 0xffff, "2cf5ff" + "00" * 0xffff)
        #self.assertEncodeDecode([None] * 0xffffff, "2df5ffff" + "00" * 0xffffff) # too slow

    def test_small_array_length(self):
        for i in range(10):
            input = [0] * i
            enc = encode_as_bytes(input)
            self.assertEqual(1 + i, len(enc))
            self.assertEqual(input, decode_bytes(enc))

        for i in range(11, 266):
            input = [0] * i
            enc = encode_as_bytes(input)
            self.assertEqual(2 + i, len(enc))
            self.assertEqual(input, decode_bytes(enc))

        for i in range(266, 0x3ff):
            input = [0] * i
            enc = encode_as_bytes(input)
            self.assertEqual(3 + i, len(enc))
            self.assertEqual(input, decode_bytes(enc))

    def test_rand_string_set(self):
        set_data = set()
        for i in range(16):
            set_data.add('k-%s' % random.randint(0, 2 ** 63))

            encoded = encode_as_bytes(set_data)
            decoded = decode_bytes(encoded)
            self.assertEqual(set(decoded), set_data)

            self.assertEqual(encoded, encode_as_bytes(decoded))

    def test_rand_object_array(self):
        data = []
        for k in range(32):
            data.append({
              'boolValue': random.random() > 0.5,
              'intValue': random.randint(-((2 ** 31) - 1), (2 ** 31) - 1),
              'longValue': random.randint(-((2 ** 63) - 1), (2 ** 63) - 1),
              'floatValue': random.random(),
            })

            encoded = encode_as_bytes(data)
            decoded = decode_bytes(encoded)
            self.assertEqual(decoded, data)


    def assertEncode(self, input_obj, expected_hex: str):
        enc = encode_as_bytes(input_obj)
        self.assertEqual(enc, bytes.fromhex(expected_hex))
        return enc

    def assertDecode(self, hex_data: str, expected_obj):
        dec = decode_bytes(bytes.fromhex(hex_data))
        self.assertEqual(dec, expected_obj)

    def assertEncodeDecode(self, input_obj, expected_hex: str):
        enc = self.assertEncode(input_obj, expected_hex)
        dec = decode_bytes(enc)
        self.assertEqual(dec, input_obj)

if __name__ == '__main__':
    unittest.main()
